;; Copyright (c) 1996 John Zuckerman
;;
;; See the file "Notice" for information on usage and redistribution
;; of this file, and for a DISCLAIMER OF ALL WARRANTIES.

(define swl:read-from-process
  (lambda (string)
    (let ([proc (process string)])
      (let ([ip (car proc)])
	(let loop ([line (swl:read-line ip)])
	  (if (not line)
	    (begin (close-input-port (car proc))
	      (close-output-port (cadr proc))
	      '())
	    (cons line (loop (swl:read-line ip)))))))))

(define swl:read-line
  (let ([op (open-output-string)])
    (lambda (ip)
      (let loop ([x (read-char ip)])
	(cond
	  [(eof-object? x) #f]
	  [(char=? x #\newline) (get-output-string op)]
	  [else (write-char x op) (loop (read-char ip))])))))

(define swl:block-read-file
  (lambda (filename consumer)
    (let ([buf (make-string 2048)])
      (let loop ([ip (open-input-file filename)])
        (let ([x (block-read ip buf 2048)])
          (unless (eof-object? x)
            (consumer (if (< x 2048) (substring buf 0 x) buf))
            (loop ip)))))))

(define swl:write-file-from-widget
  (case-lambda
    [(filename text-wgt)
     (swl:write-file-from-widget filename text-wgt '(0 . 0))]
    [(filename text-wgt mark)
     (let ([op (open-output-file filename 'replace)]
	   [buf (send text-wgt get-string mark 'end)]
	   )
       (block-write op buf (1- (string-length buf)))
       (close-output-port op))]
    ))


