;; Copyright (c) 1996 John Zuckerman
;;
;; See the file "Notice" for information on usage and redistribution
;; of this file, and for a DISCLAIMER OF ALL WARRANTIES.

(define warning-dialog
  (case-lambda
    [(parent msg) (warning-dialog parent msg '(ok))]
    [(parent msg actions)
     (define make-dialog
       (lambda (parent x y width height)
	 (let ([msg-queue (thread-make-msg-queue 'warning-dialog)])
           (let ([top (create <toplevel> with (title: "Warning"))])
	     (when parent (set-transient! top parent))
             (hide top)
             (swl:sync-display)
	     (let* ([frame (create <frame> top)]
		    [pad (create <label> frame)]
		    [msglab (create <label> frame with
			      (title: (string-append " " msg " "))
			      (background-color: 'white)
			      (border-width: 6)
			      (font: (create <font> 'helvetica 14 '(roman bold)))
			      (relief: 'groove))]
		    [bframe (create <frame> frame)]
		    [make-button
		      (lambda (sym)
			(create <button> bframe with
			  (title:
			    (string-append " " (symbol->string sym) " "))
			  (action:
			    (lambda (self)
			      (thread-send-msg msg-queue sym)))))]
		    [buttons (map make-button actions)]
		    [filler (create <label> frame with (height/char: 1))]
		    )
	       (pack frame (expand: #t) (fill: 'both))
	       (pack pad)
	       (pack msglab)
	       (pack filler)
	       (pack bframe (side: 'bottom) (fill: 'x))
	       (for-each
		 (lambda (btn) (pack btn (expand: #t) (side: 'left)))
		 buttons)
               (swl:sync-display)
               (let ([tw (send top get-width)] [th (send top get-height)])
                 (send top set-geometry!
                   (format "+~a+~a"
                     (+ x (quotient (- width tw) 2))
                     (+ y (quotient (- height th) 2))))
                 (show top)
                 (swl:sync-display))
	       (let ((msg (thread-receive-msg msg-queue)))
		 (destroy top)
		 msg))))))
     (if parent
         (swl:application-modal
           (send parent get-application-context)
           (lambda ()
             (make-dialog parent
               (get-root-x parent)
               (get-root-y parent)
               (get-width parent)
               (get-height parent))))
         (make-dialog parent 0 0 (swl:screen-width) (swl:screen-height)))]))

