;; Copyright (c) 1996 John Zuckerman
;;
;; See the file "Notice" for information on usage and redistribution
;; of this file, and for a DISCLAIMER OF ALL WARRANTIES.

(require "../common/scheme-text.ss" 'once)

(define-class (<edit-text> parent) (<scheme-text> parent)
  (ivars
    (new-menu-item #f)
    (open-menu-item #f)
    (execute-menu-item #f)
    (save-menu-item #f)
    (save-as-menu-item #f)
    (quit-menu-item #f)
    (copy-menu-item #t)
    (cut-menu-item #t)
    (paste-menu-item #t)
    (buffer-modified-menu-item #f)
    (outline-mode-menu-item #f)
    (open-menu-item-enabled? #t)
    (need-to-save-buffer? #f)
    (paste-str #f)
    )
  (inherited mini-buffer box-markup paren-markup highlight-markup
    key-prefix begin-exp end-exp end-mk handle)

  (inheritable mini-buffer box-markup paren-markup highlight-markup
    key-prefix begin-exp end-exp end-mk handle)

  (private
    [activate (mi)
      (send self turn-search-off)
      (when (send mi get-enabled) ((send mi get-action) mi))]
    )
    
  (protected)
  (public
    [get-string () (send self get-string '(0 . 0) end-mk)]

    [buffer-modified? () need-to-save-buffer?]

    [set-buffer-to-safe! () (set! need-to-save-buffer? #f)]

    [set-buffer-to-modified! ()
      (set! need-to-save-buffer? #t)
      (set-title! buffer-modified-menu-item "(Modified)")
      ]

    [delete (idx)
      (send-base self delete idx)
      (send self set-buffer-to-modified!)
      ]

    [delete (idx1 idx2)
      ;; this method is used for several things besides cutting the selection. 
      (set! need-to-save-buffer? #t)
      (send self set-buffer-to-modified!)
      (send-base self delete idx1 idx2)
      ]

    [insert-at (idx what)
      (send-base self insert-at idx what)
      (send self set-buffer-to-modified!)
      (when open-menu-item-enabled? ;;; redundant local flag is for efficiency
	(set! open-menu-item-enabled? #f)
	(send open-menu-item set-enabled! #f)
	)
      ]

    [key-press-no-prefix (key mods)
      (if (number? key) ;;; kludge test for modifier key-only
	  (send-base self key-press-no-prefix key mods)
	  (begin 
	    (event-case ((key= key) (modifier= mods))
	      [([alt #\n]) (activate new-menu-item)]
	      [([alt #\o]) (activate open-menu-item)]
	      [([alt #\g]) (activate execute-menu-item)]
	      [([alt #\s]) (activate save-menu-item)]
	      [([alt #\a]) (activate save-as-menu-item)]
	      [([alt #\q]) (activate quit-menu-item)]
	      [([alt #\c]) (activate copy-menu-item)]
	      [([alt #\x]) (activate cut-menu-item)]
	      [([alt #\v]) (activate paste-menu-item)]
	      [([control #\d])
	       (send self action-copy) ;;; buffer any selection
	       ;; let the base-class do the actual delete
	       (send-base self key-press-no-prefix key mods)
	       ]
	      [else
	       (send-base self key-press-no-prefix key mods)
	       ])))]
      
    [set-menu-items! (unsaved-m new-m open-m execute-m save-m save-as-m quit-m
		       copy-m cut-m paste-m outline-mode-m)
      (set! buffer-modified-menu-item unsaved-m)
      (set! new-menu-item new-m)
      (set! open-menu-item open-m)
      (set! execute-menu-item execute-m)
      (set! save-menu-item save-m)
      (set! save-as-menu-item save-as-m)
      (set! quit-menu-item quit-m)
      (set! copy-menu-item copy-m)
      (set! cut-menu-item cut-m)
      (set! paste-menu-item paste-m)
      (set! outline-mode-menu-item outline-mode-m)
      ]

    [destroy () (when quit-menu-item (activate quit-menu-item))]
  ))
