;; Copyright (c) 1996 John Zuckerman
;;
;; See the file "Notice" for information on usage and redistribution
;; of this file, and for a DISCLAIMER OF ALL WARRANTIES.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(require "../common/scrollframe.ss")
(require "../common/arrows.ss")
(require "../common/select-file.ss")
(require "../common/auxio.ss")
(require "../common/warning-dialog.ss")
(require "../common/app.ss")
(require "../repl/repl-text.ss")

(define-class (<traffic-light>) (<command-menu-item>)
  (ivars
    (color 'green)
    (red-title #f)
    (green-title #f)
    (show-light #t))
  (inherited)
  (inheritable)
  (private
    [get-image (name)
     (call/cc
       (lambda (ret)
         (parameterize ([error-handler (lambda x (ret #f))])
           (create <photo> with
             (filename: (format "~a/~a" (getenv "SWL_ROOT") name))))))]
    [update ()
     (when show-light
       (set-title! self
         (case color
           [(red) red-title]
           [(green) green-title]
           [else "bug in repl.ss"]))
       (swl:sync-display))])
  (protected)
  (public
    [init ()
     (send-base self init)
     (let ((broken? (memq (machine-type) '(i3nt ppcnt))))
       (if broken?
           (begin
             (set! red-title "Busy")
             (set! green-title "Ready"))
           (begin
             (set! red-title (or (get-image "redlight.gif") "Busy"))
             (set! green-title (or (get-image "grnlight.gif") "Ready")))))]
    [set-color! (clr)
     (set! color clr)
     (update)]
    [toggle ()
     (set! show-light (not show-light))
     (update)
     (unless show-light (set-title! self ""))]
    [get-color () color]))
  
(define-class (<repl-toplevel> start-k end-k) (<app-toplevel>)
  (ivars)
  (inherited mini-buffer mini-search)
  (inheritable mini-buffer mini-search)
  (private)
  (protected)
  (public
   [init (start-k end-k)
    (send-base self init)
    (let () ;;; needed for defns
     (define-syntax get-bitmap
       (if (memq (machine-type) '(i3nt ppcnt))
           (syntax-rules () [(_ alt img) alt])
           (syntax-rules () [(_ alt img) img])))
     (define-syntax simple-menu-item
       (syntax-rules ()
	 ((_ str1 str2 act (argname argval) ...)
	  (create <command-menu-item> with
		  (title: (cons str1 str2))
		  (action: act)
		  (argname argval) ...
		  ))))
     (transcript-output-port #f)
     ;;
     ;; widget bindings
     ;; 
     (let* ((scrolled-frame
	     (create <scrollframe> self
		     with (default-vscroll: #t) (sticky-hscroll: #t)))
	    (history-counter
	     (create <command-menu-item> with
		     (font: (create <font> 'fixed 12 '()))
		     (background-color: 'white)
		     (foreground-color: 'black)
		     ))
	    (repl
	     (create <repl-text> scrolled-frame with
		     (background-color: 'white)
		     (foreground-color: 'black)
		     ))) ;;; widget bindings

       (send repl set-font! (create <font> 'courier 12 '()))

       (send self notify-text repl)

       ;;
       ;; action and menu-item bindings
       ;;
       (letrec
	   ([ton-menu-item 
	     (simple-menu-item
	      "Transcript On"  ""
	      (lambda (item)
		(when (transcript-output-port)
		  (error 'transcript-on "transcript port open"))
		(thread-fork
		 (lambda ()
		   (let ((name (select-file
				(current-directory)
				"Transcript On"
				self 'non-existing)))
		     (when name (send repl transcript-on name))
		     ))))
	      (enabled: #f))]
	    [toff-menu-item
	     (simple-menu-item
	      "Transcript Off"  ""
	      (lambda (item)
		(unless (transcript-output-port)
		  (error 'transcript-off
			 "transcript port not open"))
		(send repl transcript-off))
	      (enabled: #f))]
	    [apphis-menu-action
	      (lambda (item)
		(thread-fork
		 (lambda ()
		   (let ((name (select-file
				(current-directory)
				"Append To History"
				self 'existing)))
		     (if name
			 (let ((ip (open-input-file name)))
			   (let loop ((hi (send repl history-get)))
			     (let loop2 ((item ""))
			       (let ((str (swl:read-line ip)))
				 (if str
				     (if (string=? "" str)
					 (loop (cons item hi))
					 (if (string=? "" item)
					     (loop2 str)
					     (loop2 (string-append
						     item
						     "
"
						     str))))
				     (send repl history-set!
					   (if (string=? item "")
					       hi
					       (cons item hi)))
				     ))))
			   (close-port ip)
			   #t)
			 #f)
		     ))))]
	    [savhis-menu-action
	      (lambda (item)
		(let ((hi (reverse (send repl history-get))))
		  (thread-fork
		   (lambda ()
		     (let ((name
			    (select-file (current-directory)
					 "Save History As" self
					 'non-existing)))
		       (if name
			   (let ((op (open-output-file name)))
			     (for-each
			      (lambda (str)
				(display str op)
				(display #\newline op)
				(display #\newline op)
				)
			      hi)
			     (close-port op)
			     #t)
			   #f)
		       )))))]
	    [quit-menu-item
	     (simple-menu-item
	      "_Quit"         "Alt+q"
	      (lambda (item) (send self destroy)))]
	    (copy-menu-item
	     (simple-menu-item
	      "_Copy"  "Alt+c"
	      (lambda (item) (send repl action-copy))))
	    (cut-menu-item
	     (simple-menu-item
	      "Cu_t"   "Alt+x"
	      (lambda (item) (send repl action-cut))))
	    (paste-menu-item
	     (simple-menu-item
	      "_Paste" "Alt+v"
	      (lambda (item) (send repl action-paste))))
	     (search-forward-menu-item
	      (simple-menu-item
	       "Search _Forward"    "Ctrl+s"
		(lambda (item) (send repl search-forward))))
	     (search-backward-menu-item
	      (simple-menu-item
	       "Search _Backward"   "Ctrl+r"
		(lambda (item) (send repl search-backward))))
	    ) ;;; action and menu-item bindings
	 ;;
	 ;; menu bindings
	 ;;
	 (let*
	     ((menu-file
	       (create <cascade-menu-item> with
		       (title: "_File")
		       (menu:
			(create <menu>
			 (list (simple-menu-item "_Load..." ""
                                 (lambda (item)
                                   (let ([filename
                                          (swl:file-dialog
                                            "Select a file to load"
                                            'open
                                            (file-types:
                                              '(("Scheme source" ("*.ss"))))
                                            (parent: self))])
                                     (when filename
                                       (send repl insert
                                         (format "(load ~s)~%" filename))))))
                               (simple-menu-item "Load + (_cd)..." ""
                                 (lambda (item)
                                   (let ([filename
                                          (swl:file-dialog
                                            "Select a file to load"
                                            'open
                                            (file-types:
                                              '(("Scheme source" ("*.ss"))))
                                            (parent: self))])
                                     (when filename
                                       (let f ((i (fx- (string-length filename) 1)))
                                         (cond
                                           [(fx= i -1) (void)]
                                           [(char=? #\/ (string-ref filename i))
                                            (send repl insert
                                              (format "(cd ~s)~%" (substring filename 0 i)))]
                                           [else (f (fx- i 1))]))
                                       (send repl insert
                                         (format "(load ~s)~%" filename))))))
                               ton-menu-item
			       toff-menu-item
	                       (create <cascade-menu-item> with
                                 (title: "_History")
                                 (menu:
                                   (make-menu
                                     ("_Read from file" apphis-menu-action)
                                     ("_Save to file" savhis-menu-action)
                                     ("_Clear History"
                                       (lambda (item)
                                         (send repl history-clear))))))
			       quit-menu-item)
			 ))))
	      (menu-edit
	       (create <cascade-menu-item> with
		       (title: "_Edit")
		       (menu:
			(create <menu>
			 (list copy-menu-item
			       cut-menu-item
			       paste-menu-item
			       search-forward-menu-item
			       search-backward-menu-item
			       )))))
	      (menu-control 
	       (create <cascade-menu-item> with
		       (title: "Contro_l")
		       (menu:
			(create <menu>
			 (list
			  (simple-menu-item
			   "_Interrupt"    "Ctrl+c"
			   (lambda (item)
			     (send repl interrupt)))
			  )))))
	      #;(menu-templates
	       (create <cascade-menu-item> with
		       (title: "_Templates")
		       (menu:
			(create <menu>
			 (list 
			  (simple-menu-item
			   "_cond"   "Ctrl+q c" 
			   (lambda (item)
			     (send repl insert-template 'cond)))
			  (simple-menu-item
			   "_define" "Ctrl+q d"
			   (lambda (item)
			     (send repl insert-template 'define)))
			  (simple-menu-item
			   "_lambda" "Ctrl+q a"
			   (lambda (item)
			     (send repl insert-template 'lambda)))
			  (simple-menu-item
			   "l_et"    "Ctrl+q l"
			   (lambda (item)
			     (send repl insert-template 'let)))
			  )))))
	      (traffic-light (create <traffic-light>))
	      (menu-preferences 
	       (create <cascade-menu-item> with
		       (title: "P_references")
		       (menu:
                         (make-menu
                           ("_Font..."
                             (lambda (item)
                               (swl:font-dialog
                                 self
                                 (swl:font-families 'fixed)
                                 '(-8 -10 -12 -14 -16 -18 -20 -22 -24 8 10 12 14 16 18 20 22 24)
                                 '(bold normal)
                                 (lambda () (send repl get-font))
                                 (lambda (fnt)
                                   (when fnt (send repl set-font! fnt))))))
                           ("_Toggles"
                             (create <menu>
                               (list (create <check-menu-item> with
                                       (title: "The _Box")
                                       (action:
                                         (lambda (item)
                                           (send repl show-the-box))))
                                     (create <check-menu-item> with
                                       (title: "_Wrap Text")
                                       (action:
                                         (lambda (item)
                                           (let ([wrap (send repl get-wrap)])
                                             (send repl set-wrap!
                                               (if (eq? wrap 'none)
                                                   'word
                                                   'none))))))
                                     (create <check-menu-item> with
                                       (title: "Traffic _Light")
                                       (enabled: (not (memq (machine-type) '(i3nt ppcnt))))
                                       (action:
                                         (lambda (item)
                                           (send traffic-light toggle)))))))))))

;	      (menu-help
;	       (create <cascade-menu-item> with
;		 (title: "Help")
;		 (menu:
;		  (make-menu
;		   ("Key Bindings" list)
;		   ("" list)
;		   ("Ctrl is the Control key; Alt is the Alt (or Meta) key"
;		    list)
;		   ("" list)
;		   ("Ctrl+a	Move to beginning of line" list)
;		   ("Ctrl+b	Move back one character" list)
;		   ("Ctrl+d	Delete current character (or selection)" list)
;		   ("Ctrl+e	Move to end of line" list)
;		   ("Ctrl+f	Move forward one character" list)
;		   ("Ctrl+h	Backspace and remove one character" list)
;		   ("Ctrl+k	Delete rest of line" list)
;		   ("Ctrl+n	Move to next line (or next history item)" list)
;		   ("Ctrl+p	Move to prev line (or prev history item)" list)
;		   ("Ctrl+r	Search backward for string" list)
;		   ("Ctrl+s	Search forward for string" list)
;		   ("" list)
;		   ("Ctrl+c	Interrupt expression evaluation" list)
;		   ("" list)
;		   ("Alt+b	Box the enclosing (or adjacent) expression"
;		    list)
;		   ("Alt+c	Copy text to the clipboard" list)
;		   ("Alt+d	Quit this session" list)
;		   ("Alt+n	Move to next history item" list)
;		   ("Alt+p	Move to prev history item" list)
;		   ("Alt+v	Paste text from the clipboard " list)
;		   ("Alt+x	Cut text (& place on the clipboard)" list)
;		   ("Alt+<	Move to beginning of expression" list)
;		   ("Alt+>	Move to end of expression" list)
;		   ("" list)
;		   ("Esc p	History search backward (anchored)" list)
;		   ("Esc P	History search backward" list)
;		   ("Esc n	History search forward (anchored)" list)
;		   ("Esc N	History search forward" list)
;		   ("" list)
;		   ("Ctrl+q *	Insert an expression template" list)
;		   ("Tab	Indent current line" list)
;		   ("" list)
;		   ("1.4" list)
;		   ))))

;	      (menu-h-top
;	       (create <command-menu-item> with
;		       ;;(title: "|<")
;		       (title:
;                         (get-bitmap "<<"
;                           (create <bitmap> with (data: left-double-arrow))))
;		       (action: (lambda (button)
;				  (send repl history-first)))))
	      (menu-h-backward
	       (create <command-menu-item> with
		       ;;(title: "<")
		       (title:
                         (get-bitmap "<"
                           (create <bitmap> with (data: left-arrow))))
		       (action: (lambda (button)
				  (send repl history-backward)))))
	      (menu-h-forward
	       (create <command-menu-item> with
		       ;; (title: ">")
		       (title:
                         (get-bitmap ">"
                           (create <bitmap> with (data: right-arrow))))
		       (action: (lambda (button)
				  (send repl history-forward)))))
;	      (menu-h-bottom
;	       (create <command-menu-item> with
;		       ;; (title: ">|")
;		       (title:
;                         (get-bitmap ">>"
;                           (create <bitmap> with (data: right-double-arrow))))
;		       (action: (lambda (button)
;				  (send repl history-last)))))
	      ) ;;; menu bindings

	   (send repl set-menu-items!
		 toff-menu-item
		 ton-menu-item
		 copy-menu-item
		 cut-menu-item
		 paste-menu-item
		 )
	     
	   (pack scrolled-frame (expand: #t) (fill: 'both))
	   (pack repl (expand: #t) (fill: 'both))
           (set-menu! self
             (create <menu>
               (list menu-file
                     menu-edit
                     menu-control
                     #;menu-templates
                     menu-preferences
                     traffic-light
;                    menu-h-top               ; takes up too much room for our pretty traffic light :)
                     menu-h-backward
                     history-counter
                     menu-h-forward
;                    menu-h-bottom            ; takes up too much room for our pretty traffic light :)
                     (swl:help-menu)
                     (swl:application-menu))))
	   (send traffic-light toggle)
	   (send repl set-traffic-light! traffic-light)
	   (let ((token
		  (swl:begin-application (lambda () (send self destroy)))))
	     (thread-fork
	      (lambda ()
		(thread-name "Repl")
		;;; (set! myrepl repl) ;;; for debugging
		(thread-become-server!) ;;; cafe will set the reset handler
		(send repl repl-new-cafe history-counter)
		(send self destroy)
		(swl:end-application token)
		(end-k)))
	     (start-k repl))))))
     (void)]
  )
)

(define new-repl
  (case-lambda
    [() (new-repl (lambda args (void)) void)]
    [(start-k end-k)
     (swl:make-application
      (lambda ()
	(create <repl-toplevel> start-k end-k with
		(title: "repl"))))
     (void)]))

;(swl:register-application "New Repl" new-repl)

